from file_handlers.lumps.header import HeaderParser, HeaderPacker
from file_handlers.lumps.structure_parser import Struct
from file_handlers.lumps.l35 import Lump35Parser, Lump35Packer
from pathlib import Path
from time import time
import lzma

class BSPLZMA():
    def __init__(self, fp):
        self.__header = HeaderParser(fp)
        self.__fp = fp

        self.__bsp_lzma = Struct("bsp_lzma_header",
                                 (("ID", "LZMA"),
                                  ("actualSize", "INTUL"),
                                  ("lzmaSize", "INTUL"),
                                  ("properties", "RAW/5")))

    def compress(self):
        header = self.__header.parse()

        if self.test_compressed():
            return
        
        ##we construct a new file so updating offsets is not quite so difficult.
        ##new offsets are calculated 'on the fly' as follows;
        ##we create a sorted list of tuples, containing the offset, size, and lump number of each lump
        ##now we operate on each offset sequentially.
        ##every offset that follows has the delta of size and new size subtracted
        ##the fourCC becomes the uncompressed size of the lump

        ##special cases:
        ##lump 40 is compressed as a pkzip archive is (look at 7zip for this)
        ##gamelumps are individually compressed
        ##we must test to see if lumps are empty (offset, len = 0)

        offsets_list = []
        lnum = 0
        for lump in header["lump_t"]:
            offsets_list.append([lump["fileofs"], lump["filelen"], lnum])
            lnum += 1
        offsets_list = sorted(offsets_list) ##offsets are first so it sorts by them

        bspname = Path(self.__fp.name + ".COMPRESSED")

        starttime = time()
        print("Repacking..", flush=True)
        with open(bspname, "w+b") as cfp: ##w+ as truncating if exists is desired
            lump_t = [""]*64 ##initialise blank header array - to be updated later
            total_dsize = 0 ##the total size change to apply to offsets
            cfp.write(b"\x00"*1036)##reserving space at the start of the file for headers

            for item in offsets_list:
                if item[1] == 0: ##unused lump case
                    lump_t[item[2]] = {"fileofs" : 0,"filelen" : 0,"version" : 0,"fourCC" : 0}
                    
                elif item[2] != 40 and item[2] != 35:
                    size = item[1]
                    offset = item[0]
                    offset -= total_dsize
                    
                    self.__fp.seek(item[0])
                    lumpdata = lzma.compress(self.__fp.read(item[1]))
                    ########
                    ## look into using lzma stream compression, if memory issues are encountered ##
                    ########
                    lzmasize = len(lumpdata) ##lzma header size not added
                    
                    deltasize = size - lzmasize ##how much to adjust offsets by
                    total_dsize += deltasize

                    lump_t[item[2]] = {"fileofs" : offset,
                                       "filelen" : lzmasize,
                                       "version" : 0,
                                       "fourCC" : size}
                    
                    ## we construct and insert the bsp lzma header here
                    new_lzma_header = {"ID" : "LZMA",
                                  "actualSize" : size,
                                  "lzmaSize" : lzmasize,
                                  "properties" : b"\x5d\x00\x00\x00\x01"}
                    ##properties may or may not always be \x5d\x00\x00\x00\x01;
                    ##I hope they are because I don't know how to calculate them if not
                    cfp.seek(0, 2)
                    self.__bsp_lzma.pack(cfp, new_lzma_header)
                    cfp.write(lumpdata)

                        
                elif item[2] == 35: ##lump35 - gamelumps
                    ##gamelump offsets are adjusted
                    ##the length of each lump in the header remains the same
                    ##flags become 1 from 0
                    ##an extra dummy gamelump, with all fields = 0, is added
                    ##how this dummy gamelump is used is unknown to me, as no offset is set in it..
                    self.__fp.seek(header["lump_t"][35]["fileofs"])
                    l35parser = Lump35Parser(self.__fp)
                    l35packer = Lump35Packer(self.__fp)
                    l35header = l35parser.parse_header()
                    l35_dsize = 0 ##used for
                    
                    for entry in l35header["dgamelump_t"]:
                        entry["flags"] = 1
                        self.__fp.seek(entry["fileofs"])
                        if entry["fileofs"]:
                            lumpdata = lzma.compress(self.__fp.read(entry["fileofs"]))
                            cfp.seek(0, 2)
                            new_header = {"ID" : "LZMA",
                                          "actualSize" : entry["filelen"],
                                          "lzmaSize" : len(lumpdata),##lzma header size not added
                                          "properties" : b"\x5d\x00\x00\x00\x01"}
                            self.__bsp_lzma.pack(cfp, new_header)
                            cfp.write(lumpdata)
                            deltasize += entry["filelen"] - len(lumpdata)
                            l35_dsize += deltasize
                                       
                        entry["fileofs"] -= total_dsize
                        total_dsize += deltasize
                    ##dummy header
                    l35header["dgamelump_t"].append({"ID" : b"\x00\x00\x00\x00",
                                                         "flags" : 0,
                                                         "fileofs" : 0,
                                                         "filelen" : 0})
                    l35header["lumpCount"] += 1

                    lump_t[item[2]] = {"fileofs" : item[0] - total_dsize,
                                         "filelen" : item[1] - l35_dsize,
                                         "version" : 0,
                                         "fourCC" : item[0]}
                    cfp.seek(lump_t[item[2]])
                    l35packer.pack_header(l35header, delta_ngamelumps = 1)
                    ##new l35 header with updated offsets.
                    ##for the purpose of delta_ngamelumps, see pack_header in l35.py for an explanation                    

                else: ##lump 40 - pakfile
                    lump_t[item[2]] = {"fileofs" : 0,"filelen" : 0,"version" : 0,"fourCC" : 0}
                    pass
                
                ##end of header adjustments

            cfp.seek(0) ##create the bsp header
            new_header = {"ident" : "VBSP",
                          "version" : 20,
                          "lump_t" : lump_t,
                          "revision" : header["revision"]}
            header_packer = HeaderPacker(cfp)
            header_packer.pack(new_header)

            endtime = time()

            print("\n\nRepacking finished in %s seconds" %round(endtime-starttime, 2))
            
            

    def test_compressed(self):
        data = self.__header.parse()
        return not(data["lump_t"][0]["fourCC"] == 0)


